"""babylab-redcap

A Python interface to the for the SJD Babylab REDCap database API.


This package provides convenient wrappers to interact with the NeuroDevCo Babylab's REDCap API to consult, modify or delete records from participants, appointments, or language questionnsires.

Modules:
    api: Contains the main functions used to interact with the REDCap API.
    utils: Utils functions for cleaning data, and for reformatting labels and values.
    globals: Global variables used throughout the `api` and `utiils` modules.
"""  # pylint: disable=line-too-long
